use std::{
    fmt::Display,
    thread,
    error::Error,
    sync::{
        mpsc,
        Arc,
        Mutex,
    },
};

pub struct ThreadPool {
    workers: Vec<Worker>,
    sender: Option<mpsc::Sender<Job>>,
}

type Job = Box<dyn FnOnce() + Send + 'static>;

struct Worker {
    id: usize,
    thread: Option<thread::JoinHandle<()>>,
}

impl Worker {
    fn new(id: usize, receiver: Arc<Mutex<mpsc::Receiver<Job>>>) -> Worker {
        Worker { id, thread: Some(thread::spawn(move || {
            loop {
                match (*receiver).lock().unwrap().recv() {
                    Ok(job) => {
                        println!("Worker {} got a job... running.", id);
                        job();
                    },
                    Err(_) => {
                        println!("Worker {} stopping.", id);
                        break;
                    },
                }
            }
        }))}
    }
}

#[derive(Debug)]
pub enum ThreadPoolError {
    ZeroSize
}

impl Error for ThreadPoolError {
    fn description(&self) -> &str {
        match *self {
            ThreadPoolError::ZeroSize => "ZeroPoolSize",
        }
    }
}

impl Display for ThreadPoolError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match *self {
            ThreadPoolError::ZeroSize =>
                f.write_str("Thread pool size is zero"),
        }
    }
}

impl ThreadPool {
    pub fn new(size: usize) -> Result<ThreadPool, ThreadPoolError> {
        if size == 0 {
            return Err(ThreadPoolError::ZeroSize);
        }

        let mut workers: Vec<Worker> = Vec::with_capacity(size);
        let (sender, receiver) = mpsc::channel();

        let receiver = Arc::new(Mutex::new(receiver));

        for i in 0..size {
            workers.push(Worker::new(i, Arc::clone(&receiver)));
        }

        Ok(ThreadPool { workers, sender: Some(sender) })
    }

    pub fn execute<F>(&self, f: F)
    where
        F: FnOnce() + Send + 'static
    {
        let job = Box::new(f);
        (*self).sender.as_ref().unwrap().send(job).unwrap();
    }
}

impl Drop for ThreadPool {
    fn drop(&mut self) {
        drop((*self).sender.take());

        for worker in &mut (*self).workers {
            println!("Shutting down worker {}.", worker.id);
            if let Some(thread) = worker.thread.take() {
                thread.join().unwrap();
            }
        }
    }
}
